import svgwrite.shapes

from ._mixins import MarkersMixin, PresentationMixin, TransformMixin
from .base import BaseElement

__all__ = [
    "Line",
    "Rect",
    "Circle",
    "Ellipse",
    "Polyline",
    "Polygon",
]


class Line(
    BaseElement[svgwrite.shapes.Line],
    TransformMixin,
    PresentationMixin,
    MarkersMixin,
):
    _api_name = "line"


class Rect(
    BaseElement[svgwrite.shapes.Rect], TransformMixin, PresentationMixin
):
    _api_name = "rect"


class Circle(
    BaseElement[svgwrite.shapes.Circle], TransformMixin, PresentationMixin
):
    _api_name = "circle"


class Ellipse(
    BaseElement[svgwrite.shapes.Ellipse], TransformMixin, PresentationMixin
):
    _api_name = "ellipse"


class Polyline(
    BaseElement[svgwrite.shapes.Polyline],
    TransformMixin,
    PresentationMixin,
    MarkersMixin,
):
    _api_name = "polyline"


class Polygon(
    BaseElement[svgwrite.shapes.Polygon],
    TransformMixin,
    PresentationMixin,
    MarkersMixin,
):
    _api_name = "polygon"



from abc import ABC

from ..core import BaseGlyph

__all__ = [
    "HArrayGlyph",
    "VArrayGlyph",
]

from .matrix import BaseMatrixGlyph


class BaseArrayGlyph(BaseMatrixGlyph, ABC):
    """
    Glyph encapsulating an array of glyphs with constant spacing between them,
    either horizontal or vertical.

    Horizontal arrays grow to the right, and vertical arrays grow downwards.

    If `center` is `True`, glyphs are center aligned with respect to the other
    axis. That is, horizontal arrays would be centered vertically and vertical
    arrays would be centered horizontally.
    """

    _horizontal: bool = False

    @classmethod
    def new(
        cls,
        glyphs: list[BaseGlyph],
        glyph_id: str | None = None,
        spacing: float = 0.0,
        padding: float = 0.0,
        center: bool = True,
    ):
        rows = [glyphs] if cls._horizontal else [[g] for g in glyphs]

        params = cls.get_params_cls()(
            rows=rows, spacing=spacing, padding=padding, center=center
        )
        return cls(params=params, glyph_id=glyph_id)


class HArrayGlyph(BaseArrayGlyph):
    _horizontal = True


class VArrayGlyph(BaseArrayGlyph):
    _horizontal = False



"""
Mixins to provide APIs.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Self

import svgwrite.mixins

from ..._utils import normalize_str

if TYPE_CHECKING:
    from .gradients import BaseGradient

__all__ = [
    "ViewBoxMixin",
    "TransformMixin",
    "PresentationMixin",
    "MarkersMixin",
]


def _normalize_color(
    color: str | None, gradient: BaseGradient | None
) -> str | None:
    return (
        color
        if color is not None
        else gradient.funciri
        if gradient is not None
        else None
    )


class BaseWrapperMixin[T]:
    """
    Wraps an svgwrite mixin.
    """

    _mixin_obj: T


class ViewBoxMixin(BaseWrapperMixin[svgwrite.mixins.ViewBox]):
    def viewbox(self, min_x: float, min_y: float, width: float, height: float):
        self._mixin_obj.viewbox(min_x, min_y, width, height)

    def stretch(self):
        self._mixin_obj.stretch()

    def fit(
        self,
        horiz: Literal["left", "center", "right"] = "center",
        vert: Literal["top", "middle", "bottom"] = "middle",
        scale: Literal["meet", "slice"] = "meet",
    ):
        self._mixin_obj.fit(horiz=horiz, vert=vert, scale=scale)


class TransformMixin(BaseWrapperMixin[svgwrite.mixins.Transform]):
    def translate(self, x: float, y: float | None = None) -> Self:
        self._mixin_obj.translate(x, y)
        return self

    def rotate(
        self, angle: float, center: tuple[float, float] | None = None
    ) -> Self:
        # set center if none was provided and we have a size
        if center is None:
            if (size := self._mixin_size) is not None:
                center = (size[0] / 2, size[1] / 2)

        self._mixin_obj.rotate(angle, center=center)
        return self

    def scale(self, x: float, y: float | None = None) -> Self:
        self._mixin_obj.scale(x, y)
        return self

    def skew_x(self, angle: float) -> Self:
        self._mixin_obj.skewX(angle)
        return self

    def skew_y(self, angle: float) -> Self:
        self._mixin_obj.skewY(angle)
        return self

    def matrix(
        self,
        scale_x: float,
        scale_y: float,
        skew_x: float,
        skew_y: float,
        translate_x: float,
        translate_y: float,
    ) -> Self:
        self._mixin_obj.matrix(
            scale_x, scale_y, skew_x, skew_y, translate_x, translate_y
        )
        return self

    @property
    def _mixin_size(self) -> tuple[float, float] | None:
        """
        Overridden by subclass if applicable.
        """
        return None


class PresentationMixin(BaseWrapperMixin[svgwrite.mixins.Presentation]):
    def fill(
        self,
        color: str | None = None,
        gradient: BaseGradient | None = None,
        rule: str | None = None,
        opacity: float | None = None,
    ):
        self._mixin_obj.fill(
            color=_normalize_color(color, gradient),
            rule=rule,
            opacity=normalize_str(opacity),
        )

    def stroke(
        self,
        color: str | None,
        gradient: BaseGradient | None = None,
        width: float | None = None,
        opacity: float | None = None,
        linecap: Literal["butt", "round", "square"] | None = None,
        linejoin: Literal["arcs", "bevel", "miter", "miter-clip", "round"]
        | None = None,
        miterlimit: int | float | None = None,
    ):
        self._mixin_obj.stroke(
            color=_normalize_color(color, gradient),
            width=normalize_str(width),
            opacity=normalize_str(opacity),
            linecap=normalize_str(linecap),
            linejoin=normalize_str(linejoin),
            miterlimit=normalize_str(miterlimit),
        )

    def dasharray(
        self,
        dasharray: list[int | float] | None = None,
        offset: float | str | None = None,
    ):
        self._mixin_obj.dasharray(
            dasharray=dasharray, offset=normalize_str(offset)
        )


class MarkersMixin(BaseWrapperMixin[svgwrite.mixins.Markers]):
    pass


# may not be necessary
"""
class ClippingMixin(BaseWrapperMixin[svgwrite.mixins.Clipping]):
    pass
"""



from ..core import BaseGlyph, BaseParams

__all__ = [
    "MatrixParams",
    "MatrixGlyph",
]


class MatrixParams(BaseParams):
    rows: list[list[BaseGlyph]]
    spacing: float = 0.0
    padding: float = 0.0
    center: bool = True


class BaseMatrixGlyph(BaseGlyph[MatrixParams]):
    """
    Base matrix class, used for matrix glyph and array glyphs.
    """

    _rows: list[list[BaseGlyph]]
    _cols: list[list[BaseGlyph]]

    _max_width: float
    _max_height: float

    def init(self):
        # validate rows
        for i, row in enumerate(self.params.rows):
            assert len(row) == len(
                self.params.rows[i - 1]
            ), f"Row lengths inconsistent: {self.params.rows}"

        self.size_canon = self._get_size()

    def draw(self):
        if len(self.params.rows) == 0:
            return

        for row_idx, row in enumerate(self._rows):
            for col_idx, glyph in enumerate(row):
                insert_x: float
                insert_y: float

                # set insert point
                insert_x = col_idx * (self._max_width + self.params.spacing)
                insert_y = row_idx * (self._max_height + self.params.spacing)

                # adjust insert point if centered
                if self.params.center:
                    insert_x += (self._max_width - glyph.size[0]) / 2
                    insert_y += (self._max_height - glyph.size[1]) / 2

                # add padding
                insert_x += self.params.padding
                insert_y += self.params.padding

                # insert glyph
                self.insert_glyph(glyph, (insert_x, insert_y))

    def _get_size(self) -> tuple[float, float]:
        """
        Get the size of this matrix based on the sizes of the glyphs.
        """

        width: float
        height: float

        if len(self.params.rows) == 0:
            return (0.0, 0.0)

        rows: list[list[BaseGlyph]] = list(
            [list(row) for row in self.params.rows]
        )
        cols: list[list[BaseGlyph]] = list(map(list, zip(*rows)))

        # get column widths
        col_widths = [max([g.size[0] for g in col]) for col in cols]

        # get row heights
        row_heights = [max([g.size[1] for g in row]) for row in rows]

        # set rows/cols
        self._rows = rows
        self._cols = cols

        # get max width/height
        self._max_width = max(col_widths)
        self._max_height = max(row_heights)

        # get total width
        width = sum(col_widths) + self.params.spacing * (len(cols) - 1)

        # get total height
        height = sum(row_heights) + self.params.spacing * (len(rows) - 1)

        # add padding
        width += self.params.padding * 2
        height += self.params.padding * 2

        return (width, height)


class MatrixGlyph(BaseMatrixGlyph):
    """
    Glyph encapsulating a matrix of glyphs with constant spacing between them
    and padding around the edges.

    If `center` is `True`, glyphs are center aligned.
    """

    @classmethod
    def new(
        cls,
        rows: list[list[BaseGlyph]],
        glyph_id: str | None = None,
        spacing: float = 0.0,
        padding: float = 0.0,
        center: bool = True,
    ):
        params = cls.get_params_cls()(
            rows=rows, spacing=spacing, padding=padding, center=center
        )
        return cls(params=params, glyph_id=glyph_id)

    @property
    def rows(self) -> list[list[BaseGlyph]]:
        return self._rows

    @property
    def cols(self) -> list[list[BaseGlyph]]:
        return self._cols



"""
Exports classes which encapsulate an SVG element.
"""

from pyrollup import rollup

from . import base, containers, gradients, shapes
from .base import *  # noqa
from .containers import *  # noqa
from .gradients import *  # noqa
from .shapes import *  # noqa

__all__ = rollup(base, containers, shapes, gradients)



from pyrollup import rollup

from . import export, glyph, graphics
from .export import *  # noqa
from .glyph import *  # noqa
from .graphics import *  # noqa

__all__ = rollup(glyph, graphics, export)



from __future__ import annotations

from typing import Any, TypeVar, get_args


def extract_type_param[
    T, U
](cls: type[T], param_cls: type[U]) -> type[U] | None:
    """
    Extract the type param with the given class.
    """

    orig_bases = cls.__orig_bases__  # pyright: ignore[reportGeneralTypeIssues]
    assert len(orig_bases)

    args = get_args(orig_bases[0])
    if len(args):
        param: type[U]

        if isinstance(args[0], TypeVar):
            # have a TypeVar, look up its bound
            type_var = args[0]
            assert type_var.__bound__ is not None
            param = type_var.__bound__
        else:
            # already have a class
            param = args[0]

        assert issubclass(param, param_cls)
        return param

    return None


def normalize_str(v: Any | None) -> str | None:
    return None if v is None else str(v)



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING

import svgwrite.base

from ..._utils import normalize_str
from ..properties import ShapeProperties
from .base import BaseElement
from .gradients import LinearGradient, RadialGradient
from .shapes import Circle, Ellipse, Line, Polygon, Polyline, Rect

if TYPE_CHECKING:
    from .base import BaseGlyph
    from .containers import Group

__all__ = [
    "ElementFactory",
]


class ElementFactory(ABC):
    @property
    @abstractmethod
    def _glyph(self) -> BaseGlyph:
        ...

    @property
    @abstractmethod
    def _container(self) -> svgwrite.base.BaseElement:
        ...

    def draw_line(
        self,
        start: tuple[float, float],
        end: tuple[float, float],
        properties: ShapeProperties | None = None,
    ) -> Line:
        return Line(
            self._glyph,
            self._container,
            start=start,
            end=end,
            **self._get_extra(properties),
        )

    def draw_polyline(
        self,
        points: Iterable[tuple[float, float]],
        properties: ShapeProperties | None = None,
    ) -> Polyline:
        return Polyline(
            self._glyph,
            self._container,
            points=[p for p in points],
            **self._get_extra(properties),
        )

    def draw_polygon(
        self,
        points: Iterable[tuple[float, float]],
        properties: ShapeProperties | None = None,
    ) -> Polygon:
        return Polygon(
            self._glyph,
            self._container,
            points=[p for p in points],
            **self._get_extra(properties),
        )

    def draw_rect(
        self,
        insert: tuple[float, float],
        size: tuple[float, float],
        radius_x: float | None = None,
        radius_y: float | None = None,
        properties: ShapeProperties | None = None,
    ) -> Rect:
        return Rect(
            self._glyph,
            self._container,
            insert=insert,
            size=size,
            rx=radius_x,
            ry=radius_y,
            **self._get_extra(properties),
        )

    def draw_circle(
        self,
        center: tuple[float, float],
        radius: float,
        properties: ShapeProperties | None = None,
    ) -> Circle:
        return Circle(
            self._glyph,
            self._container,
            center=center,
            r=radius,
            **self._get_extra(properties),
        )

    def draw_ellipse(
        self,
        center: tuple[float, float],
        radius: tuple[float, float],
        properties: ShapeProperties | None = None,
    ) -> Ellipse:
        return Ellipse(
            self._glyph,
            self._container,
            center=center,
            r=radius,
            **self._get_extra(properties),
        )

    def create_group(self, properties: ShapeProperties | None = None) -> Group:
        from .containers import Group

        # only use passed properties instead of inheriting from glyph
        extra = properties._get_values() if properties else {}

        return Group(self._glyph, self._container, **extra)

    def create_linear_gradient(
        self,
        start: tuple[float, float] | None = None,
        end: tuple[float, float] | None = None,
        inherit: str | BaseElement | None = None,
        colors: list[str] | None = None,
        sweep_pct: tuple[float] = (0.0, 100.0),
        opacity: float | None = None,
    ) -> LinearGradient:
        elem = LinearGradient(
            self._glyph,
            self._glyph._svg.defs,
            start=start,
            end=end,
            inherit=_normalize_inherit(inherit),
            gradientUnits="userSpaceOnUse",
        )

        if colors:
            elem._element.add_colors(
                colors,
                sweep=[s / 100 for s in sweep_pct],
                opacity=normalize_str(opacity),
            )

        return elem

    def create_radial_gradient(
        self,
        center: tuple[float, float] | None = None,
        radius: float | None = None,
        focal: tuple[float, float] | None = None,
        inherit: str | BaseElement | None = None,
        colors: list[str] | None = None,
        sweep_pct: tuple[float] = (0.0, 100.0),
        opacity: float | None = None,
    ) -> RadialGradient:
        elem = RadialGradient(
            self._glyph,
            self._glyph._svg.defs,
            center=center,
            r=radius,
            focal=focal,
            inherit=_normalize_inherit(inherit),
            gradientUnits="userSpaceOnUse",
        )

        if colors:
            elem._element.add_colors(
                colors,
                sweep=[s / 100 for s in sweep_pct],
                opacity=normalize_str(opacity),
            )

        return elem

    def _get_extra(self, properties: ShapeProperties | None) -> dict[str, str]:
        """
        Get extra kwargs to pass to svgwrite APIs.
        """
        # aggregate provided properties with those of the glyph
        props = ShapeProperties._aggregate(
            [self._glyph.properties] + ([properties] if properties else [])
        )
        return props._get_values()


def _normalize_inherit(inherit: str | BaseElement | None) -> str | None:
    if isinstance(inherit, BaseElement):
        return inherit.iri



from __future__ import annotations

from typing import TYPE_CHECKING

import svgwrite.container

from ._factory import ElementFactory
from ._mixins import PresentationMixin, TransformMixin
from .base import BaseElement

if TYPE_CHECKING:
    from ...glyph import BaseGlyph

__all__ = [
    "Group",
]


class Group(
    BaseElement[svgwrite.container.Group],
    TransformMixin,
    PresentationMixin,
    ElementFactory,
):
    _api_name = "g"

    @property
    def _glyph(self) -> BaseGlyph:
        return self._glyph_obj

    @property
    def _container(self) -> svgwrite.container.Group:
        return self._element



from __future__ import annotations

from typing import Any, Self, cast

from pydantic import BaseModel, ConfigDict

__all__ = [
    "PropertyValueType",
    "Properties",
    "ShapeProperties",
    "GradientProperties",
]

type PropertyValueType = str | None


class BasePropertiesModel(BaseModel):
    """
    Encapsulates graphics properties, as defined here:
    <https://www.w3.org/TR/SVG11/intro.html#TermProperty>

    And listed here: <https://www.w3.org/TR/SVG11/propidx.html>
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def _aggregate(cls, models: list[BasePropertiesModel]) -> Self:
        """
        Create a new model with values aggregated from provided models.
        """
        values: dict[str, str] = {}

        for model in models:
            values.update(model._get_values())

        return cls(**values)

    def _get_values(self) -> dict[str, str]:
        values: dict[str, str] = {}

        for field in list(self.model_fields.keys()):
            value = cast(Any | None, getattr(self, field))
            if isinstance(value, str):
                values[field] = value

        return values


# TODO: use appropriate type, e.g. opacity should be numeric


class ColorPropertiesMixin(BasePropertiesModel):
    """
    Common color-related properties.
    """

    color: PropertyValueType = None
    color_interpolation: PropertyValueType = None
    color_interpolation_filters: PropertyValueType = None
    color_profile: PropertyValueType = None
    color_rendering: PropertyValueType = None
    opacity: PropertyValueType = None


class PaintingPropertiesMixin(ColorPropertiesMixin, BasePropertiesModel):
    """
    Properties related to painting operations.
    """

    fill: PropertyValueType = None
    fill_opacity: PropertyValueType = None
    fill_rule: PropertyValueType = None
    marker: PropertyValueType = None
    marker_end: PropertyValueType = None
    marker_mid: PropertyValueType = None
    marker_start: PropertyValueType = None
    stroke: PropertyValueType = None
    stroke_dasharray: PropertyValueType = None
    stroke_dashoffset: PropertyValueType = None
    stroke_linecap: PropertyValueType = None
    stroke_linejoin: PropertyValueType = None
    stroke_miterlimit: PropertyValueType = None
    stroke_opacity: PropertyValueType = None
    stroke_width: PropertyValueType = None
    shape_rendering: PropertyValueType = None


class FontPropertiesMixin(BasePropertiesModel):
    """
    Properties related to font specification.
    """

    font: PropertyValueType = None
    font_family: PropertyValueType = None
    font_size: PropertyValueType = None
    font_size_adjust: PropertyValueType = None
    font_stretch: PropertyValueType = None
    font_style: PropertyValueType = None
    font_variant: PropertyValueType = None
    font_weight: PropertyValueType = None


class TextPropertiesMixin(ColorPropertiesMixin, BasePropertiesModel):
    """
    Properties related to text layout and rendering.
    """

    direction: PropertyValueType = None
    letter_spacing: PropertyValueType = None
    text_decoration: PropertyValueType = None
    unicode_bidi: PropertyValueType = None
    word_spacing: PropertyValueType = None
    writing_mode: PropertyValueType = None
    alignment_baseline: PropertyValueType = None
    baseline_shift: PropertyValueType = None
    dominant_baseline: PropertyValueType = None
    glyph_orientation_horizontal: PropertyValueType = None
    glyph_orientation_vertical: PropertyValueType = None
    kerning: PropertyValueType = None
    text_anchor: PropertyValueType = None
    text_rendering: PropertyValueType = None


class ImagePropertiesMixin(BasePropertiesModel):
    """
    Properties specific to image elements.
    """

    # Specific to image rendering
    image_rendering: PropertyValueType = None
    preserve_aspect_ratio: PropertyValueType = None


class ClippingMaskingPropertiesMixin(BasePropertiesModel):
    """
    Properties related to clipping and masking.
    """

    clip: PropertyValueType = None
    clip_path: PropertyValueType = None
    clip_rule: PropertyValueType = None
    mask: PropertyValueType = None


class GradientPropertiesMixin(ColorPropertiesMixin, BasePropertiesModel):
    """
    Properties specific to gradients.
    """

    stop_color: PropertyValueType = None
    stop_opacity: PropertyValueType = None


class FilterEffectPropertiesMixin(ColorPropertiesMixin, BasePropertiesModel):
    """
    Properties related to filter effects.
    """

    enable_background: PropertyValueType = None
    filter: PropertyValueType = None
    flood_color: PropertyValueType = None
    flood_opacity: PropertyValueType = None
    lighting_color: PropertyValueType = None


class CursorPropertiesMixin(BasePropertiesModel):
    """
    Properties related to cursors.
    """

    cursor: PropertyValueType = None
    pointer_events: PropertyValueType = None


class ViewportPropertiesMixin(BasePropertiesModel):
    """
    Properties related to the viewport.
    """

    overflow: PropertyValueType = None
    display: PropertyValueType = None
    visibility: PropertyValueType = None


class Properties(
    PaintingPropertiesMixin,
    FontPropertiesMixin,
    TextPropertiesMixin,
    ClippingMaskingPropertiesMixin,
    GradientPropertiesMixin,
    FilterEffectPropertiesMixin,
    CursorPropertiesMixin,
    ViewportPropertiesMixin,
    BasePropertiesModel,
):
    """
    Class to represent all styling properties:
    <https://www.w3.org/TR/SVG11/styling.html#SVGStylingProperties>
    """

    def __init_subclass__(cls):
        super().__init_subclass__()

        valid_properties = Properties.model_fields.keys()

        # ensure user didn't add any invalid properties
        for field in cls.model_fields.keys():
            assert field in valid_properties, f"{field} is not a valid property"


class ShapeProperties(
    PaintingPropertiesMixin,
    FilterEffectPropertiesMixin,
    CursorPropertiesMixin,
    ViewportPropertiesMixin,
    BasePropertiesModel,
):
    """
    Properties applicable to basic shapes.
    """


class GradientProperties(
    GradientPropertiesMixin,
    ColorPropertiesMixin,
):
    """
    Properties applicable to gradients.
    """



import os

from pyrollup import rollup

from . import elements, properties
from .elements import *  # noqa
from .properties import *  # noqa

__all__ = ["RASTER_SUPPORT"] + rollup(elements, properties)

RASTER_SUPPORT: bool = os.name == "posix"
"""
Whether rasterization is supported; Linux only.
"""



import svgwrite.gradients

from .base import BaseElement

__all__ = [
    "BaseGradient",
    "LinearGradient",
    "RadialGradient",
]


class BaseGradient[GradientT: svgwrite.gradients._AbstractGradient](
    BaseElement[GradientT]
):
    def get_paint_server(self) -> str:
        return self._element.get_paint_server()


class LinearGradient(BaseGradient[svgwrite.gradients.LinearGradient]):
    _api_name = "linearGradient"


class RadialGradient(BaseGradient[svgwrite.gradients.RadialGradient]):
    _api_name = "radialGradient"



"""
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Callable, cast

import svgwrite.base

from ._mixins import BaseWrapperMixin

if TYPE_CHECKING:
    from ...glyph import BaseGlyph

__all__ = [
    "BaseElement",
]


class BaseElement[ElementT: svgwrite.base.BaseElement](BaseWrapperMixin):
    """
    Wraps an `svgwrite` SVG element and corresponding API mixins.
    """

    _api_name: str
    """
    Name of svgwrite API to invoke.
    """

    _element: ElementT
    """
    Instance of svgwrite element.
    """

    _glyph_obj: BaseGlyph

    def __init__(
        self,
        glyph: BaseGlyph,
        container: svgwrite.base.BaseElement,
        *args,
        **kwargs,
    ):
        """
        Creates the corresponding `svgwrite` element, passing through
        extra kwargs. Should not be instantiated directly; use draw APIs.
        """

        self._glyph_obj = glyph

        api_attr = getattr(glyph._drawing, self._api_name)
        api = cast(Callable[..., ElementT], api_attr)

        self._element = api(*args, **kwargs)
        self._mixin_obj = self._element

        container.add(self._element)

    @property
    def iri(self) -> str:
        return self._element.get_iri()

    @property
    def funciri(self) -> str:
        return self._element.get_funciri()
